from typing import Optional, Union
from uuid import UUID

from deprecation import deprecated

from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine.informatics.workflows.workflow import Workflow
from citrine.resources.design_execution import DesignExecutionCollection
from citrine._rest.ai_resource_metadata import AIResourceMetadata

__all__ = ['DesignWorkflow']


class DesignWorkflow(Resource['DesignWorkflow'], Workflow, AIResourceMetadata):
    """Object that generates scored materials that may approach higher values of the score.

    Parameters
    ----------
    name: str
        the name of the workflow
    design_space_id: Optional[UUID]
        the UUID corresponding to the design space to use
    predictor_id: Optional[UUID]
        the UUID corresponding to the predictor to use
    predictor_version: Optional[Union[int, str]]
        the version of the predictor to use
    description: Optional[str]
        a description of the workflow

    """

    design_space_id = properties.Optional(properties.UUID, 'design_space_id')
    predictor_id = properties.Optional(properties.UUID, 'predictor_id')
    predictor_version = properties.Optional(
        properties.Union([properties.Integer(), properties.String()]), 'predictor_version')
    _branch_id: Optional[UUID] = properties.Optional(properties.UUID, 'branch_id')
    """:Optional[UUID]: Unique ID of the branch that contains this workflow."""

    status_description = properties.String('status_description', serializable=False)
    """:str: more detailed description of the workflow's status"""
    typ = properties.String('type', default='DesignWorkflow', deserializable=False)

    _branch_root_id: Optional[UUID] = properties.Optional(properties.UUID, 'branch_root_id',
                                                          serializable=False, deserializable=False)
    """:Optional[UUID]: Root ID of the branch that contains this workflow."""
    _branch_version: Optional[int] = properties.Optional(properties.Integer, 'branch_version',
                                                         serializable=False, deserializable=False)
    """:Optional[int]: Version number of the branch that contains this workflow."""

    def __init__(self,
                 name: str,
                 *,
                 design_space_id: Optional[UUID] = None,
                 predictor_id: Optional[UUID] = None,
                 predictor_version: Optional[Union[int, str]] = None,
                 description: Optional[str] = None):
        self.name = name
        self.design_space_id = design_space_id
        self.predictor_id = predictor_id
        self.predictor_version = predictor_version
        self.description = description

    def __str__(self):
        return '<DesignWorkflow {!r}>'.format(self.name)

    @property
    def design_executions(self) -> DesignExecutionCollection:
        """Return a resource representing all visible executions of this workflow."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot initialize execution without project reference!')
        return DesignExecutionCollection(
            project_id=self.project_id, session=self._session, workflow_id=self.uid)

    @property
    @deprecated(deprecated_in="2.42.0", removed_in="3.0.0",
                details="Please use the branch_root_id and branch_version instead.")
    def branch_id(self):
        """[deprecated] Retrieve the version ID of the branch this workflow is on."""
        return self._branch_id

    @branch_id.setter
    @deprecated(deprecated_in="2.42.0", removed_in="3.0.0",
                details="Please set the branch_root_id and branch_version instead.")
    def branch_id(self, value):
        self._branch_id = value
        self._branch_root_id = None
        self._branch_version = None

    @property
    def branch_root_id(self):
        """Retrieve the root ID of the branch this workflow is on."""
        return self._branch_root_id

    @branch_root_id.setter
    def branch_root_id(self, value):
        self._branch_root_id = value
        self._branch_id = None

    @property
    def branch_version(self):
        """Retrieve the version of the branch this workflow is on."""
        return self._branch_version

    @branch_version.setter
    def branch_version(self, value):
        self._branch_version = value
        self._branch_id = None



import warnings
from copy import deepcopy
from typing import Callable, Union, Iterable, Optional, Tuple
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._session import Session
from citrine.exceptions import NotFound
from citrine.informatics.workflows import DesignWorkflow
from citrine.resources.response import Response
from functools import partial


class DesignWorkflowCollection(Collection[DesignWorkflow]):
    """A collection of DesignWorkflows."""

    _path_template = '/projects/{project_id}/design-workflows'
    _individual_key = None
    _collection_key = 'response'
    _resource = DesignWorkflow
    _api_version = "v2"

    def __init__(self,
                 project_id: UUID,
                 session: Session,
                 branch_id: Optional[UUID] = None,
                 *,
                 branch_root_id: Optional[UUID] = None,
                 branch_version: Optional[int] = None):
        self.project_id: UUID = project_id
        self.session: Session = session

        if branch_id:
            warnings.warn("Constructing the design workflow interface by its parent branch's "
                          "version ID is deprecated. Please use its root ID and version number.",
                          DeprecationWarning)
            if branch_root_id:
                raise ValueError("Please provide at most one: the version ID or root ID.")

        self.branch_id = branch_id
        self.branch_root_id = branch_root_id
        self.branch_version = branch_version

    def _resolve_branch_root_and_version(self, workflow):
        from citrine.resources.branch import BranchCollection

        workflow_copy = deepcopy(workflow)
        bc = BranchCollection(self.project_id, self.session)
        branch = bc.get_by_version_id(version_id=workflow_copy._branch_id)
        workflow_copy._branch_root_id = branch.root_id
        workflow_copy._branch_version = branch.version
        return workflow_copy

    def _resolve_branch_id(self, root_id, version):
        from citrine.resources.branch import BranchCollection

        if root_id and version:
            bc = BranchCollection(self.project_id, self.session)
            branch = bc.get(root_id=root_id, version=version)
            return branch.uid
        return None

    def register(self, model: DesignWorkflow) -> DesignWorkflow:
        """
        Upload a new design workflow.

        The model's branch ID is ignored in favor of this collection's. If this
        collection has a null branch ID, then a branch is first created, then
        the design workflow is created on it.

        Parameters
        ----------
        model: DesignWorkflow
            The design workflow to be uploaded.

        Return
        -------
        DesignWorkflow
            The newly created design workflow.

        """
        if self.branch_root_id is None or self.branch_version is None:
            if self.branch_id is None:
                # There are a number of contexts in which hitting design workflow endpoints without
                # a branch ID is valid, so only this particular usage is deprecated.
                msg = ('A design workflow must be created with a branch. Please use '
                       'branch.design_workflows.register() instead of '
                       'project.design_workflows.register().')
                raise RuntimeError(msg)
            else:
                warnings.warn("Registering a design workflow to a branch by its parent's version "
                              "ID is deprecated. Please use its root ID and version number.",
                              DeprecationWarning)
                branch_id = self.branch_id
        else:
            # branch_id is in the body of design workflow endpoints, so it must be serialized.
            # This means the collection branch_id might not match the workflow branch_id. The
            # collection should win out, since the user is explicitly referencing the branch
            # represented by this collection.
            # To avoid modifying the parameter, and to ensure the only change is the branch_id, we
            # deepcopy, modify, then register it.
            branch_id = self._resolve_branch_id(self.branch_root_id, self.branch_version)
        model_copy = deepcopy(model)
        model_copy._branch_id = branch_id
        return super().register(model_copy)

    def build(self, data: dict) -> DesignWorkflow:
        """
        Build an individual design workflow from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the design workflow.

        Return
        -------
        DesignWorkflow
            The design workflow created from data.

        """
        workflow = DesignWorkflow.build(data)
        workflow = self._resolve_branch_root_and_version(workflow)
        workflow._session = self.session
        workflow.project_id = self.project_id
        return workflow

    def update(self, model: DesignWorkflow) -> DesignWorkflow:
        """Update a design workflow.

        Identifies the workflow by the model's uid. It must have a branch_id, and if this
        collection also has a branch_id, they must match. Prefer updating a workflow through
        Project.design_workflows.update.

        Parameters
        ----------
        model: DesignWorkflow
            The design workflow values that are desired, identified by the uid.

        Return
        -------
        DesignWorkflow
            The design workflow resulting from the update.

        """
        if self.branch_root_id is not None or self.branch_version is not None:
            if self.branch_root_id != model.branch_root_id or \
                    self.branch_version != model.branch_version:
                raise ValueError('To move a design workflow to another branch, please use '
                                 'Project.design_workflows.update')
        elif self.branch_id is not None:
            warnings.warn("Updating a design workflow by its parent's branch version ID is "
                          "deprecated. Please use its root ID and version number.",
                          DeprecationWarning)
            if self.branch_id != model._branch_id:
                raise ValueError('To move a design workflow to another branch, please use '
                                 'Project.design_workflows.update')

        if model.branch_root_id is not None and model.branch_version is not None:
            try:
                model._branch_id = self._resolve_branch_id(model.branch_root_id,
                                                           model.branch_version)
            except NotFound:
                raise ValueError('Cannot update a design workflow unless its branch_root_id and '
                                 'branch_version exists.')
        elif model._branch_id is None:
            raise ValueError('Cannot update a design workflow unless its branch_root_id and '
                             'branch_version are set.')

        # If executions have already been done, warn about future behavior change
        executions = model.design_executions.list()
        if next(executions, None) is not None:
            raise RuntimeError("Cannot update a design workflow after candidate generation, "
                               "please register a new design workflow instead")

        return super().update(model)

    def archive(self, uid: Union[UUID, str]):
        """Archive a design workflow.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the workflow to archive

        """
        url = self._get_path(uid=uid, action="archive")
        self.session.put_resource(url, {}, version=self._api_version)

    def restore(self, uid: Union[UUID, str]):
        """Restore an archived design workflow.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the workflow to restore

        """
        url = self._get_path(uid=uid, action="restore")
        self.session.put_resource(url, {}, version=self._api_version)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """Design Workflows cannot be deleted; they can be archived instead."""
        raise NotImplementedError(
            "Design Workflows cannot be deleted; they can be archived instead.")

    def list_archived(self, *, per_page: int = 500) -> Iterable[DesignWorkflow]:
        """List archived Design Workflows."""
        fetcher = partial(self._fetch_page, additional_params={"filter": "archived eq 'true'"})
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page)

    def _fetch_page(self,
                    path: Optional[str] = None,
                    fetch_func: Optional[Callable[..., dict]] = None,
                    page: Optional[int] = None,
                    per_page: Optional[int] = None,
                    json_body: Optional[dict] = None,
                    additional_params: Optional[dict] = None,
                    ) -> Tuple[Iterable[dict], str]:
        params = additional_params or {}
        params["branch"] = self._resolve_branch_id(self.branch_root_id, self.branch_version)
        return super()._fetch_page(path=path,
                                   fetch_func=fetch_func,
                                   page=page,
                                   per_page=per_page,
                                   json_body=json_body,
                                   additional_params=params)



from typing import Optional, Type

from citrine.informatics.predictors import GraphPredictor
from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine.informatics.catalyst.language_model import LanguageModelChoice


class AssistantRequest(Serializable["AssistantRequest"]):
    """A query to the model assistant."""

    question = properties.String("question")
    predictor = properties.Object(GraphPredictor, "config")
    temperature = properties.Optional(properties.Float, "temperature", default=0.0)
    language_model = properties.Optional(properties.Enumeration(LanguageModelChoice),
                                         "language_model", default=LanguageModelChoice.GPT_4)

    def __init__(self, *,
                 question: str,
                 predictor: GraphPredictor,
                 temperature: Optional[float] = 0.0,
                 language_model: Optional[LanguageModelChoice] = LanguageModelChoice.GPT_4):
        self.question = question
        self.predictor = predictor
        self.temperature = temperature
        self.language_model = language_model

    def _post_dump(self, data: dict) -> dict:
        # When dumping, GraphPredictor.dump isn't called, resulting in some fields being dropped.
        data["config"] = self.predictor.dump()["instance"]
        return super()._post_dump(data)


class AssistantResponse(PolymorphicSerializable["AssistantResponse"]):
    """The parent type for all Model Assistant responses."""

    @classmethod
    def get_type(cls, data) -> Type['AssistantResponse']:
        """Return the subtype."""
        type_dict = {
            "message": AssistantResponseMessage,
            "modified-config": AssistantResponseConfig,
            "unsupported": AssistantResponseUnsupported,
            "input-error": AssistantResponseInputErrors,
            "exec-error": AssistantResponseExecError
        }
        typ = type_dict.get(data['type'])
        if typ is not None:
            return typ
        else:
            raise ValueError(
                f'{data["type"]} is not a valid assistant response type. '
                f'Must be in {type_dict.keys()}.'
            )


class AssistantResponseMessage(Serializable["AssistantResponseMessage"], AssistantResponse):
    """A successful model assistant invocation, whose response is only text."""

    message = properties.String("data.message")


class AssistantResponseConfig(Serializable["AssistantResponseConfig"], AssistantResponse):
    """A successful model assistant invocation, whose response includes a modified predictor."""

    predictor = properties.Object(GraphPredictor, "data.config")

    @classmethod
    def _pre_build(cls, data):
        data["data"]["config"] = GraphPredictor.wrap_instance(data["data"]["config"])
        return data


class AssistantResponseUnsupported(Serializable["AssistantResponseUnsupported"],
                                   AssistantResponse):
    """A successful model assistant invocation, but for an unsupported query.

    This will cover any user query which the model assistant could not map to a functionality it
    supports. For example, asking a generic question, or something unrelated to materials.
    """

    message = properties.String("data.message")


class AssistantResponseInputError(Serializable["AssistantResponseInputError"], AssistantResponse):
    """A single input failure.

    Contains the error message, and the field it applies to.
    """

    field = properties.Optional(properties.String, "field")
    error = properties.String("error")


class AssistantResponseInputErrors(Serializable["AssistantResponseInputErrors"],
                                   AssistantResponse):
    """A failed model assistant invocation, due to malformed input.

    This should only happen if there's some field omitted by the client, or one of its values is
    outside acceptable ranges.
    """

    errors = properties.List(properties.Object(AssistantResponseInputError), "data.errors")


class AssistantResponseExecError(Serializable["AssistantResponseExecError"], AssistantResponse):
    """A failed model assistant invocation, due to some internal issue.

    This most likely indicates the assistant got some unexpected output when asking its query. It
    indicates an internal error, not a problem with the client invocation.
    """

    error = properties.String("data.error")



from citrine.informatics.catalyst.insights import InsightsResponse, InsightsRequest
from citrine.informatics.catalyst.assistant import AssistantResponse, AssistantRequest
from citrine.informatics.predictors.graph_predictor import GraphPredictor
from citrine.resources.user import UserCollection
from citrine._session import Session
from citrine._utils.functions import resource_path


class CatalystResource:
    """Encapsulates th ability to invoke Catalyst."""

    _path_template: str = '/catalyst'
    _api_version = 'v1'

    def __init__(self, session: Session):
        self.session: Session = session

    def _get_path(self, action: str):
        """Construct a Catalyst url from a base path and action."""
        return resource_path(path_template=self._path_template, action=action)

    def _ensure_internal_user(self):
        """Ensure that the user is internal; raise an error otherwise."""
        user = UserCollection(self.session).me()
        if not user.is_internal:
            raise NotImplementedError(
                "I'm sorry, but this feature is not currently available for "
                "your organization. Please watch the release notes for "
                "updates."
            )

    def assistant(self, query: str, *, predictor: GraphPredictor) -> AssistantResponse:
        """Invoke the model assistant.

        Parameters
        ----------
        query: str
            The query or instruction to pass to the assistant
        predictor: GraphPredictor
            The predictor you wish for the assistant to consider or act on in your query

        Returns
        -------
        The assistant response, containing details of the result of your query which vary by type.

        """
        self._ensure_internal_user()

        payload = AssistantRequest(question=query, predictor=predictor).dump()
        path = self._get_path("assistant")
        data = self.session.post_resource(path, json=payload, version=self._api_version)
        return AssistantResponse.build(data)

    def insights(self, query: str) -> InsightsResponse:
        """Invoke the insights.

        Parameters
        ----------
        query: str
            The query or instruction to pass to the insights

        Returns
        -------
        The insights response, containing details of the result of your query which vary by type.

        """
        self._ensure_internal_user()

        payload = InsightsRequest(question=query).dump()
        path = self._get_path(["documents", "search"])
        data = self.session.post_resource(path, json=payload, version=self._api_version)
        return InsightsResponse.build(data)



import functools
import warnings
from typing import Iterator, Optional, Union
from uuid import UUID

from deprecation import deprecated

from citrine._rest.collection import Collection
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import migrate_deprecated_argument
from citrine.exceptions import NotFound
from citrine.resources.data_version_update import BranchDataUpdate, NextBranchVersionRequest
from citrine.resources.design_workflow import DesignWorkflowCollection
from citrine.resources.experiment_datasource import (ExperimentDataSourceCollection,
                                                     ExperimentDataSource)


LATEST_VER = "latest"  # Refers to the most recently created branch version.


class Branch(Resource['Branch']):
    """
    A project branch.

    A branch is a container for design workflows.
    """

    name = properties.String('data.name')
    uid = properties.Optional(properties.UUID(), 'id')
    archived = properties.Boolean('metadata.archived', serializable=False)
    created_at = properties.Optional(properties.Datetime(), 'metadata.created.time',
                                     serializable=False)
    updated_at = properties.Optional(properties.Datetime(), 'metadata.updated.time',
                                     serializable=False)
    # added in v2
    root_id = properties.UUID('metadata.root_id', serializable=False)
    version = properties.Integer('metadata.version', serializable=False)

    project_id: Optional[UUID] = None

    def __init__(self,
                 name: str,
                 *,
                 session: Optional[Session] = None):
        self.name: str = name
        self.session: Session = session

    def __str__(self):
        return f'<Branch {self.name!r}>'

    @property
    def design_workflows(self) -> DesignWorkflowCollection:
        """Return a resource representing all workflows contained within this branch."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot initialize workflow without project reference!')
        return DesignWorkflowCollection(project_id=self.project_id,
                                        session=self.session,
                                        branch_root_id=self.root_id,
                                        branch_version=self.version)

    @property
    def experiment_datasource(self) -> Optional[ExperimentDataSource]:
        """Return this branch's experiment data source, or None if one doesn't exist."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot retrieve datasource without project reference!')
        erds = ExperimentDataSourceCollection(project_id=self.project_id, session=self.session)
        branch_erds_iter = erds.list(branch_version_id=self.uid, version=LATEST_VER)
        return next(branch_erds_iter, None)

    def _post_dump(self, data: dict) -> dict:
        # Only the data portion of an entity is sent to the server.
        data = data["data"]
        return data


class BranchCollection(Collection[Branch]):
    """A collection of Branches."""

    _path_template = '/projects/{project_id}/branches'
    _individual_key = None
    _collection_key = 'response'
    _resource = Branch
    _api_version = 'v2'

    def __init__(self, project_id: UUID, session: Session):
        self.project_id: UUID = project_id
        self.session: Session = session

    def build(self, data: dict) -> Branch:
        """
        Build an individual branch from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the branch.

        Return
        -------
        Branch
            The branch created from data.

        """
        branch = Branch.build(data)
        branch.session = self.session
        branch.project_id = self.project_id
        return branch

    @deprecated(deprecated_in="2.42.0", removed_in="3.0.0", details="Use .get() instead.")
    def get_by_root_id(self,
                       *,
                       branch_root_id: Union[UUID, str],
                       version: Optional[Union[int, str]] = LATEST_VER) -> Branch:
        """
        Given a branch root ID and (optionally) the version, retrieve the specific branch version.

        Parameters
        ---------
        branch_root_id: Union[UUID, str]
            Unique identifier of the branch root

        version: Union[int, str], optional
            The version of the branch to retrieve. If provided, must either be a positive integer,
            or "latest". Defaults to "latest".

        Returns
        -------
        Branch
            The requested version of the branch.

        """
        return self.get(root_id=branch_root_id, version=version)

    def get(self,
            uid: Optional[Union[UUID, str]] = None,
            *,
            root_id: Optional[Union[UUID, str]] = None,
            version: Optional[Union[int, str]] = LATEST_VER) -> Branch:
        """
        Retrieve a branch using either the version ID, or the root ID and version number.

        Providing both the version ID and the root ID, or neither, will result in an error.

        Providing the root ID and no version number will retrieve the latest version.

        Using the version ID with this method is deprecated in favor of .get_by_version_id().

        Parameters
        ---------
        uid: Union[UUID, str], optional
            [deprecated] Unqiue ID of the branch version.

        root_id: Union[UUID, str], optional
            Unique identifier of the branch root

        version: Union[int, str], optional
            The version of the branch to retrieve. If provided, must either be a positive integer,
            or "latest". Defaults to "latest".

        Returns
        -------
        Branch
            The requested version of the branch.

        """
        if uid:
            warnings.warn("Retrieving a branch by its version ID using .get() is deprecated. "
                          "Please use .get_by_version_id().",
                          DeprecationWarning)

            if root_id:
                raise ValueError("Please provide exactly one: the version ID or root ID.")
            return self.get_by_version_id(version_id=uid)
        elif root_id:
            version = version or LATEST_VER
            params = {"root": str(root_id), "version": version}
            branch = next(self._list_with_params(per_page=1, **params), None)
            if branch:
                return branch
            else:
                raise NotFound.build(
                    message=f"Branch root '{root_id}', version {version} not found",
                    method="GET",
                    path=self._get_path(),
                    params=params
                )

        else:
            raise ValueError("Please provide exactly one: the version ID or root ID.")

    def get_by_version_id(self, *, version_id: Union[UUID, str]) -> Branch:
        """
        Given a branch version ID, retrieve the branch.

        Parameters
        ---------
        version_id: Union[UUID, str]
            Unique ID of the branch version

        Returns
        -------
        Branch
            The requested branch.

        """
        return super().get(version_id)

    def list(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List all branches using pagination.

        Yields all branches, regardless of archive status, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            All branches in this collection.

        """
        warnings.warn("Beginning in the 3.0 release, this method will only list unarchived "
                      "branches. To list all branches, use .list_all().",
                      DeprecationWarning)
        return super().list(per_page=per_page)

    def list_archived(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List archived branches using pagination.

        Yields all archived branches, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            Archived branches in this collection.

        """
        return self._list_with_params(per_page=per_page, archived=True)

    def list_all(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List active branches using pagination.

        Yields all active branches, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            Active branches in this collection.

        """
        return self._list_with_params(per_page=per_page)

    def _list_with_params(self, *, per_page, **kwargs):
        fetcher = functools.partial(self._fetch_page, additional_params=kwargs)
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page)

    def archive(self,
                uid: Optional[Union[UUID, str]] = None,
                *,
                root_id: Optional[Union[UUID, str]] = None,
                version: Optional[Union[int, str]] = LATEST_VER):
        """
        Archive a branch.

        Parameters
        ----------
        uid: Union[UUID, str], optional
            [deprecated] Unique identifier of the branch

        root_id: Union[UUID, str], optional
            Unique ID of the branch root

        version: Union[int, str], optional
            The version of the branch. If provided, must either be a positive integer, or "latest".
            Defaults to "latest".

        """
        if uid:
            warnings.warn("Archiving a branch by its version ID is deprecated. "
                          "Please use its root ID and version number.",
                          DeprecationWarning)
        elif root_id:
            version = version or LATEST_VER
            # The backend API at present expects a branch version ID, so we must look it up.
            uid = self.get(root_id=root_id, version=version).uid
        else:
            raise ValueError("Please provide exactly one: the version ID or root ID.")

        url = self._get_path(uid, action="archive")
        data = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(data)

    def restore(self,
                uid: Optional[Union[UUID, str]] = None,
                *,
                root_id: Optional[Union[UUID, str]] = None,
                version: Optional[int] = None):
        """
        Restore an archived branch.

        Parameters
        ----------
        uid: Union[UUID, str], optional
            [deprecated] Unique identifier of the branch

        root_id: Union[UUID, str], optional
            Unique ID of the branch root

        version: Union[int, str], optional
            The version of the branch. If provided, must either be a positive integer, or "latest".
            Defaults to "latest".

        """
        if uid:
            warnings.warn("Restoring a branch by its version ID is deprecated. "
                          "Please use its root ID and version number.",
                          DeprecationWarning)
        elif root_id:
            version = version or LATEST_VER
            # The backend API at present expects a branch version ID, so we must look it up.
            uid = self.get(root_id=root_id, version=version).uid
        else:
            raise ValueError("Please provide exactly one: the version ID or root ID.")

        url = self._get_path(uid, action="restore")
        data = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(data)

    def update_data(self,
                    branch: Optional[Union[UUID, str, Branch]] = None,
                    *,
                    root_id: Optional[Union[UUID, str]] = None,
                    version: Optional[Union[int, str]] = LATEST_VER,
                    use_existing: bool = True,
                    retrain_models: bool = False) -> Optional[Branch]:
        """
        Automatically advance the branch to the next version.

        If there are no newer versions of data sources used by this branch this method returns
        without doing anything

        Parameters
        ----------
        branch: Union[UUID, str, Branch], optional
            [deprecated] Branch Identifier or Branch object

        root_id: Union[UUID, str], optional
            Unique ID of the branch root

        version: Union[int, str], optional
            The version of the branch. If provided, must either be a positive integer, or "latest".
            Defaults to "latest".

        use_existing: bool
            If true the workflows in this branch will use existing predictors that are using
            the latest versions of the data sources and are ready to use.

        retrain_models: bool
            If true, when new versions of models are created, they are automatically
            scheduled for training.

        Returns
        -------
        Branch
            The new branch record after version update or None if no update

        """
        if branch:
            if root_id:
                raise ValueError("Please provide exactly one: the version ID or root ID.")

            if isinstance(branch, Branch):
                warnings.warn("Updating a branch's data by its object is deprecated. "
                              "Please use its root ID and version number.",
                              DeprecationWarning)
            else:
                warnings.warn("Updating a branch's data by its version ID is deprecated. "
                              "Please use its root ID and version number.",
                              DeprecationWarning)
                branch = self.get_by_version_id(version_id=branch)
            root_id = branch.root_id
            version = branch.version
        elif not root_id:
            raise ValueError("Please provide exactly one: the version ID or root ID.")

        version = version or LATEST_VER
        version_updates = self.data_updates(root_id=root_id, version=version)
        # If no new data sources, then exit, nothing to do
        if len(version_updates.data_updates) == 0:
            return None

        use_predictors = []
        if use_existing:
            use_predictors = version_updates.predictors

        branch_instructions = NextBranchVersionRequest(data_updates=version_updates.data_updates,
                                                       use_predictors=use_predictors)
        branch = self.next_version(root_id=root_id,
                                   branch_instructions=branch_instructions,
                                   retrain_models=retrain_models)
        return branch

    def data_updates(self,
                     uid: Optional[Union[UUID, str]] = None,
                     *,
                     root_id: Optional[Union[UUID, str]] = None,
                     version: Optional[Union[int, str]] = LATEST_VER) -> BranchDataUpdate:
        """
        Get data updates for a branch.

        Parameters
        ----------
        uid: Union[UUID, str], optional
            [deprecated] Unqiue ID of the branch version.

        root_id: Union[UUID, str], optional
            Unique ID of the branch root

        version: Union[int, str], optional
            The version of the branch. If provided, must either be a positive integer, or "latest".
            Defaults to "latest".

        Returns
        -------
        BranchDataUpdate
            A list of data updates and compatible predictors

        """
        if uid:
            warnings.warn("Retrieving data updates for a branch by its version ID is deprecated. "
                          "Please use its root ID and version number.",
                          DeprecationWarning)
            if root_id:
                raise ValueError("Please provide exactly one: the version ID or root ID.")
        elif root_id:
            version = version or LATEST_VER
            # The backend API at present expects a branch version ID, so we must look it up.
            uid = self.get(root_id=root_id, version=version).uid
        else:
            raise ValueError("Please provide exactly one: the version ID or root ID.")

        path = self._get_path(uid=uid, action="data-version-updates-predictor")
        data = self.session.get_resource(path, version=self._api_version)
        return BranchDataUpdate.build(data)

    def next_version(self,
                     branch_root_id: Optional[Union[UUID, str]] = None,
                     root_id: Optional[Union[UUID, str]] = None,
                     *,
                     branch_instructions: NextBranchVersionRequest,
                     retrain_models: bool = True):
        """
        Move a branch to the next version.

        Parameters
        ----------
        branch_root_id: Union[UUID, str], optional
            [deprecated] Unique identifier of the branch root to advance to next version.
            Deprecated in favor of root_id.

        root_id: Union[UUID, str], optional
            Unique identifier of the branch root to advance to next version

        branch_instructions: NextBranchVersionRequest
            Instructions for how the next version of a branch should handle its predictors
            when the workflows are cloned.  data_updates contains the list of data source
            versions to upgrade (current->latest), and use_predictors will either have a
            <predictor_id>:latest to indicate the workflow should use a new version of
            the predictor.  Or <predictor_id>:<version #> to indicate that the workflow
            should use an existing predictor version.

        retrain_models: bool
            If true, when new versions of models are created, they are automatically
            scheduled for training

        Returns
        -------
        Branch
            The new branch record after version update

        """
        root_id = migrate_deprecated_argument(root_id, "root_id", branch_root_id, "branch_root_id")

        path = self._get_path(action="next-version-predictor")
        data = self.session.post_resource(path, branch_instructions.dump(),
                                          version=self._api_version,
                                          params={
                                              'root': str(root_id),
                                              'retrain_models': retrain_models})
        return self.build(data)
